import math
import struct
import numpy as np
import matplotlib          as mpl
import matplotlib.pyplot   as plt
import matplotlib.patches  as ptc
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

######################################################
#
#    Calculation of multiplicity of cryo EM views 
#    by L. Urzhumtseva & A.Urzhumtsev, February 23, 2023
#    version 2.3
# 
######################################################

def InputData(filedata,filemes): 

    nfdat = open(filedata, 'r')
    nfmes = open(filemes, 'w')

#   default values of the parameters updatable by KeyWords

    AngGrid, AngEps, AngMax                     =     90,   0.01,   3.0
    ViewFreq, ViewWeight                        = 'cell',   'inv'
    ShowAxis                                    =    0.0,    0.0 

    SizeX, ShiftH, ShowDots                     =    8.0,    0.0,   0.002 
    DrawHist, DrawMap, DrawView                 =  'yes',  'yes',  'yes'
    PrintFlag                                   =      1

    ColorBottom, ColorTop, ColorNum             =  0.125,    8.0,    256
    ColorHist, ColorPlus, ColorMap, ColorViews  = 'blue',  'red',  'jet' ,     0
    FileProj, FileProjW, FileType, FileStruct   = 'unkn', 'unkn',  'unkn', 'unkn'
    HistNum, HistBottom, HistTop                = 'unkn', 'unkn',  'unkn'

#   technical parameters for drawing in module DrawAll (not updatable by KeyWords) :

#   position of the ColorBar depending on the plots chosen
#   Number of labels in the ColorBar
#   position of the main title
#   parameters of the histogram plot
#   define the label size as a part of the figure size

    ColorBox1   = [0.005, 0.12, 0.01, 0.75]
    ColorBox2   = [0.05 , 0.12, 0.01, 0.75]
    ColorBox4   = [0.05 , 0.53, 0.01, 0.35]
    NumBarTicks = 6
    TitleX      = 0.95
    HistBarWidth , HistBarShift           = 0.4, 0.3
    SizeLabMin, SizeLabMax, SizeLabScale  =   6,  24, 1.25

    print('============== input data ==============',file=nfmes)

#   evential data modification

    ktype =   0

    while (True) :
      line = nfdat.readline()
      LineCont = line.split()
      kline = len(LineCont)

      if kline == 0 :
         break
      elif LineCont[0][0] == '#' :
         continue
      elif kline <= 2 :
         print('*** Incomplete line in the file of parameters ***',file=nfmes)
         print(line,file=nfmes)
         exit()
      elif LineCont[1] != '=' :
         print('*** Missed assignement symbol = in the file of parameters ***',file=nfmes)
         print(line,file=nfmes)
         exit()

      elif LineCont[0] == 'FileProj' or LineCont[0] == 'FP':
         FileProj = LineCont[2].rstrip('\n')
         print('FileName for the input file with projections    =',file=nfmes)
         print('  ',FileProj,file=nfmes)
      elif LineCont[0] == 'FileProjW' or LineCont[0] == 'FW':
         FileProjW = LineCont[2].rstrip('\n')
         print('Output file with projections and weights',5*' ',' =',file=nfmes)
         print('  ',FileProjW,file=nfmes)
      elif LineCont[0] == 'FileType' or LineCont[0] == 'FT':
         FileType = LineCont[2]
         ktype = 1
         print('Imposed FileType for the input projections file = ',FileType,file=nfmes)
      elif LineCont[0] == 'FileStruct' or LineCont[0] == 'FS':
         FileStruct = LineCont[2]
         print('Imposed Keyword to detect projections data      = ',FileStruct,file=nfmes)

      elif LineCont[0] == 'AngGrid' or LineCont[0] == 'AG':
         AngGrid = int(LineCont[2])
         print('Updated AngularGrid Number',18*' ','  = ',AngGrid,file=nfmes)
      elif LineCont[0] == 'AngMax' or LineCont[0] == 'AM':
         AngMax = float(LineCont[2])
         print('Updated Angular radius of a view',12*' ','  = ',AngMax,file=nfmes)
      elif LineCont[0] == 'AngEps' or LineCont[0] == 'AE':
         AngEps = float(LineCont[2])
         print('Updated Contribution Accuracy value',9*' ','  = ',AngEps,file=nfmes)
      elif LineCont[0] == 'ViewFreq' or LineCont[0] == 'VF':
         ViewFreq = LineCont[2]
         print('Updated ViewsFrequency calculation Method',4*' ',' = ',ViewFreq,file=nfmes)
      elif LineCont[0] == 'ViewWeigth' or LineCont[0] == 'VW':
         ViewWeigth = LineCont[2]
         print('Updated View-Weight calculation Method',6*' ','  = ',ViewWeigth,file=nfmes)
      elif LineCont[0] == 'ShowAxis' or LineCont[0] == 'SA':
         kphi = 3
         if kline >= 5 and (LineCont[3] == ',' or LineCont[3] == ';') :
            LineCont[3] = LineCont[4]
         if kline <= 3 :
            print('*** Incomplete line in the file of parameters ***',file=nfmes)
            print(line,file=nfmes)
            exit()
         ksym = len(LineCont[2])
         ShowAxis = float(LineCont[2]),float(LineCont[3])
         print('Updated Show direction for the Views plot',4*' ',' = ',ShowAxis,file=nfmes)

      elif LineCont[0] == 'DrawHist' or LineCont[0] == 'DH':
         DrawHist = LineCont[2]
         print('Updated DrawHistogram Flag',7*' ',' = ',DrawHist,file=nfmes)
      elif LineCont[0] == 'DrawMap' or LineCont[0] == 'DM':
         DrawMap  = LineCont[2]
         print('Updated DrawFrequencyMap Flag',16*' ',' = ',DrawMap,file=nfmes)
      elif LineCont[0] == 'DrawView' or LineCont[0] == 'DV':
         DrawView = LineCont[2]
         print('Updated DrawViews Flag',23*' ',' = ',DrawView,file=nfmes)

      elif LineCont[0] == 'SizeX' or LineCont[0] == 'SX':
         SizeX = float(LineCont[2])
         print('Updated X-size of each figure',16*' ',' = ',SizeX,file=nfmes)
      elif LineCont[0] == 'ShiftH' or LineCont[0] == 'SH':
         ShiftH = float(LineCont[2])
         print('Updated Horizontal Shift between figures',5*' ',' = ',ShiftH,file=nfmes)
      elif LineCont[0] == 'ShowDots' or LineCont[0] == 'SD':
         ShowDots = float(LineCont[2])
         print('Updated Dots Radius at the Views plot',8*' ',' = ',ShowDots,file=nfmes)

      elif LineCont[0] == 'HistNum' or LineCont[0] == 'HN':
         HistNum = int(LineCont[2])
         print('Updated Angular radius of a view',13*' ',' = ',HistNum,file=nfmes)
      elif LineCont[0] == 'HistBottom' or LineCont[0] == 'HB':
         HistBottom = float(LineCont[2])
         print('Updated Histogram Bottom value',14*' ','  = ',HistBottom,file=nfmes)
      elif LineCont[0] == 'HistTop' or LineCont[0] == 'HT':
         HistTop = float(LineCont[2])
         print('Updated Histogram Top value',18*' ',' = ',HistTop,file=nfmes)

      elif LineCont[0] == 'ColorBottom' or LineCont[0] == 'CB':
         ColorBottom = float(LineCont[2])
         print('Updated Color Bottom value',19*' ',' = ',ColorBottom,file=nfmes)
      elif LineCont[0] == 'ColorTop' or LineCont[0] == 'CT':
         ColorTop = float(LineCont[2])
         print('Updated MaxLogValue for colors',15*' ',' = ',ColorTop,file=nfmes)
      elif LineCont[0] == 'ColorNum' or LineCont[0] == 'CN':
         ColorNum = int(LineCont[2])
         print('Updated Color Number for Interpolation',7*' ',' = ',ColorNum,file=nfmes)

      elif LineCont[0] == 'ColorHist' or LineCont[0] == 'CH':
         ColorHist = LineCont[2]
         print('Updated Color for the Histogram plot',9*' ',' = ',ColorHist,file=nfmes)
      elif LineCont[0] == 'ColorPlus' or LineCont[0] == 'CP':
         ColorPlus = LineCont[2]
         print('Updated Color for the Positive-values Histogram = ',ColorPlus,file=nfmes)
      elif LineCont[0] == 'ColorMap' or LineCont[0] == 'CM':
         ColorMap = LineCont[2]
         print('Updated Color for the Map of views',11*' ',' = ',ColorMap,file=nfmes)
      elif LineCont[0] == 'ColorViews' or LineCont[0] == 'CV':
         ColorViews = int(LineCont[2])
         print('Updated Color for the Views',7*' ',' = ',ColorViews,file=nfmes)
      elif LineCont[0] == 'PrintFlag' or LineCont[0] == 'PF':
         PrintFlag = int(LineCont[2])
         if PrintFlag < 0 :
            PrintFlag = 0
         elif PrintFlag > 2 :
            PrintFlag = 2
         print('Updated PrintFlag',28*' ',' = ',PrintFlag,file=nfmes)

      else :      
         print('*** Unknown keyword in the file of parameters = ',file=nfmes)
         print(LineCont[0],file=nfmes)
         exit()

#   data processing and return

    if FileProj == 'unkn' :
       print('*** Input file with projectins is not defined ***')
       exit()

    if AngGrid <= 0 :
       print('*** Wrong AngGrid number = ',AngGrid,file=nfmes)
       exit()
    if AngEps <= 0.0 or AngEps >= 1.0 :
       print('*** Wrong Contribution Accuracy = ',AngEps,file=nfmes)
       exit()
    if AngMax <= 0.0 or AngMax >= 45.0 :
       print('*** Wrong Angular radius = ',AngMax,file=nfmes)
       exit()
    if SizeX <= 0.0 :
       print('*** Wrong Figure X-size = ',SizeX,file=nfmes)
       exit()
    if ShowDots <= 0.0 or ShowDots > 0.1:
       print('*** Wrong Dots Radius for the Views plot = ',ShowDots,file=nfmes)
       exit()
    if HistNum != 'unkn' and HistNum <= 0 :
       print('*** Wrong Histogram Number of bins = ',HistNum,file=nfmes)
       exit()
    if HistTop != 'unkn' and HistBottom != 'unkn' and HistTop <= HistBottom :
       print('*** Wrong Histogram Bottom/Top bounds = ',HistBottom,HistTop,file=nfmes)
       exit()
    if ColorBottom <= 0.0 :
       print('*** Wrong Color Bottom value = ',ColorBottom,file=nfmes)
       exit()
    if ColorTop <= ColorBottom :
       print('*** Wrong Color Bottom/Top values = ',ColorBottom,ColorTop,file=nfmes)
       exit()
    if ColorNum <= 1 :
       print('*** Wrong Color Number for interpolation = ',ColorNum,file=nfmes)
       exit()

    if ViewFreq == 'Cell' or ViewFreq == 'CELL' or ViewFreq == 'cell' :
       AngFreq = 1
    elif ViewFreq == 'Int' or ViewFreq == 'INT'  or ViewFreq == 'int': 
       AngFreq = 2
    else :
       print('*** Wrong ViewsFrequency calculation Method = ',ViewFreq,file=nfmes)
       exit()

    if ViewWeight == 'Inv' or ViewWeight == 'INV' or ViewWeight == 'inv' :
       AngWeight = 1
    elif ViewWeight == 'Log' or ViewWeight == 'LOG' or ViewWeight == 'log': 
       AngWeight = 2
    else :
       print('*** Wrong View-Weights Method = ',ViewWeight,file=nfmes)
       exit()

    print(' ',file=nfmes)
    if DrawHist == 'No' or DrawHist == 'NO' or DrawHist == 'no' \
                        or DrawHist == 'N'  or DrawHist == 'n' :
       print('Updated Views Histogram',22*' ',' =  do not plot',file=nfmes)
       DrawHist = 'no'
    elif DrawHist == 'Yes' or DrawHist == 'YES' or DrawHist == 'yes' \
                           or DrawHist == 'Y'   or DrawHist == 'y':
       print('Updated Views Histogram',22*' ',' =  plot',file=nfmes)
       DrawHist = 'yes'
    else :
       print('*** Wrong flag to plot histograms = ',DrawHist,file=nfmes)
       exit()

    if DrawMap == 'No' or DrawMap == 'NO' or DrawMap == 'no' \
                       or DrawMap == 'N'  or DrawMap == 'n' :
       print('Updated Frequency Map',24*' ',' =  do not plot',file=nfmes)
       DrawMap  = 'no'
    elif DrawMap == 'Yes' or DrawMap == 'YES' or DrawMap == 'yes'\
                          or DrawMap == 'Y'   or DrawMap == 'y':
       print('Updated Frequency Map',24*' ',' =  plot',file=nfmes)
       DrawMap  = 'yes'
    else :
       print('*** Wrong flag to plot FrequencyMap = ',DrawMap,file=nfmes)
       exit()

    if DrawView == 'No' or DrawView == 'NO' or DrawView == 'no' \
                        or DrawView == 'N'  or DrawView == 'n' :
       print('Updated Views Map',28*' ',' =  do not plot',file=nfmes)
       DrawView  = 'no'
    elif DrawView == 'Yes' or DrawView == 'YES' or DrawView == 'yes'\
                           or DrawView == 'Y'   or DrawView == 'y':
       print('Updated Views Map',28*' ',' =  plot',file=nfmes)
       DrawView  = 'yes'
    else :
       print('*** Wrong flag to plot Individual views = ',DrawView,file=nfmes)
       exit()

    if ktype == 0 :
       FileCopy = FileProj.replace("."," ")
       LineCont = FileCopy.split()
       LenName = len(LineCont)
       FileType = LineCont[LenName-1]
       print(' ',file=nfmes)
       print('Format of the input file of projections identified as',FileType,file=nfmes)

    AngMax   = AngMax * math.pi / 180.
    AngSigma = -0.5*AngMax*AngMax / math.log(AngEps)

#   conversion cm into inches

    SizeX  = SizeX  / 2.54
    ShiftH = ShiftH / 2.54 

#   packing parameters

    ParamAuxBar  = (ColorBox1,ColorBox2,ColorBox4,NumBarTicks)
    ParamAuxHist = (HistBarWidth,HistBarShift)
    ParamAuxSize = (SizeLabMin, SizeLabMax, SizeLabScale)

    ParamCalc = (AngGrid,AngEps,AngMax,AngSigma,AngFreq,AngWeight)
    ParamFile = ((nfmes,PrintFlag,FileProj,FileProjW,FileType),FileStruct,ShowAxis)
    ParamStat = (HistNum,HistBottom,HistTop)
    ParamFig  = (DrawHist,DrawMap,DrawView,SizeX,ShiftH,ShowDots)
    ParamPlot = (ColorHist,ColorPlus,ColorMap,ColorViews,ColorBottom,ColorTop,ColorNum)
    ParamAux  = (ParamAuxBar,TitleX,ParamAuxHist,ParamAuxSize)

    return ParamCalc, ParamStat, ParamFig, ParamPlot, ParamFile, ParamAux

#============================
def GetViews(ParamFile): 

#
#   all intermediate rotation matrices are given by columns, i.e. the elements are
#   R[0] = R[0,0], R[1] = R[1,0], R[2] = R[2,0], R[3] = R[0,1], R[4] = R[1,1], ...
#
    (nfmes,PrintFlag,FileProj,FileProjW,FileType),FileStruct,ShowAxis = ParamFile

    if   FileType == 'test' :
         projections,FileStructN = InputProjectionsTest(FileProj,FileStruct,nfmes)
         views                   = ProjectionsViewsRelion(projections,ShowAxis,nfmes)

    elif FileType == 'star' :
         projections,FileStructN = InputProjectionsStar(FileProj,FileStruct,nfmes)
         views                   = ProjectionsViewsRelion(projections,ShowAxis,nfmes)

    elif FileType == 'hed' :
         projections,FileStructN = InputProjectionsImagic(FileProj,FileStruct,nfmes)
         views                   = ProjectionsViewsImagic(projections,ShowAxis,nfmes)

    else :
       print('UNKNOWN FILE TYPE FOR PROJECTIONS')
       exit ()

    nprojs = len(projections)

    print('',file=nfmes)
    print('============== end of input data ==============',file=nfmes)
    print('',file=nfmes)

    print(f'Total number of views',' '*20,f'{nprojs:12}', file=nfmes)

    ParamFileN = ((nfmes,PrintFlag,FileProj,FileProjW,FileType),FileStructN)

    return views, ParamFileN

#============================
def InputProjectionsTest(FileProj,FileStruct,nfmes): 

    projections = []

    nfproj = open(FileProj, 'r')

    while True:
      line = nfproj.readline()
      if not line :
         break
      LineCont = line.split()
      projections.append([LineCont[0],LineCont[1],LineCont[2]])

    nfproj.close()

    FileStructN = 0

    return projections,FileStructN
   
#============================
def InputProjectionsStar(FileProj,FileStruct,nfmes): 

    projections = []

    nfproj = open(FileProj, 'r')
#
#   block name is defined
#
    if FileStruct != 'unkn' :

       print ('*** Input projections file : Search for the keyword ',FileStruct,
              file=nfmes)

       NumTitles = FindKey1(nfproj,FileStruct)
       if NumTitles < 0 :
          print ('*** EOF reached before the data block found',file=nfmes)
          exit()

#      find the following 'loop_' record

       NumTitle = FindKey2(nfproj,'loop_')
       if NumTitle < 0 :
          print ('*** the keyword is not followed by loop_',file=nfmes)
          exit()
       else :
          NumTitles = NumTitles + NumTitle

#      read label records

       NumTitle,MaxLabel,NumPosAngles,line = GetLabels(nfproj)
       if NumTitle < 0:
          print ('*** wrong label descriptors',file=nfmes)
          exit()

       NumTitles = NumTitles + NumTitle
       if NumPosAngles[0] == 0 or NumPosAngles[1] == 0 or NumPosAngles[2] == 0 :
          print ('*** missed angles labels in the required data block ',file=nfmes)
          exit()

#      read data

       projections,NumLines,LastLine = GetDataProj(nfproj,MaxLabel,NumPosAngles,line)
       if NumLines < 0 :
          print ('*** wrong record in the list of projections',file=nfmes)
          print (LastLine,file=nfmes)
          exit()
#
#   block name is not defined
#
    else :

       print ('*** Input projections file : Search for the data block with angle labels',
              file=nfmes)

       NumTitles = 0

       while (True) :

#         read all data till 'loop_'

          NumTitle = FindKey1(nfproj,'loop_')
          if NumTitle < 0 :
             print ('*** EOF reached',file=nfmes)
             exit()

          NumTitles = NumTitles + NumTitle
  
#         read label records

          NumTitle,MaxLabel,NumPosAngles,line = GetLabels(nfproj)
          if NumTitle < 0 :
             print ('*** wrong label descriptors or EOF reached',file=nfmes)
             print (line,file=nfmes)
             exit()

          NumTitles = NumTitles + NumTitle
          if NumPosAngles[0]*NumPosAngles[0]*NumPosAngles[0] > 0 :
             break
          else :
             NumTitles = NumTitles + 1   

#      the data block with projections found; read data
 
       projections,NumLines,LastLine = GetDataProj(nfproj,MaxLabel,NumPosAngles,line)

       if NumLines < 0 :
          print ('*** wrong record in the list of projections',file=nfmes)
          print (LastLine,file=nfmes)
#
#   got all information
#
    nfproj.close()

    FileStructN = (NumTitles,MaxLabel,NumPosAngles)

    return projections,FileStructN     
                         
#============================
def InputProjectionsImagic(FileProj,FileStruct,nfmes): 

    projections = []

    nfproj = open(FileProj, 'rb')

#   detailed Imagic format (not required here)
#    form = '@14i 4s 2i 12f 80s f 14i 3f 31i 3f 2i 3f i 13f 2i 77f 228s'

#   format to extract angles, number of blocks and machine stamp
    form = '@12s i 240s 3f 4s i 748s'

    nbytes = 1024

    while True:
      line = nfproj.read(nbytes)
      if not line :
         break
      LineCont = struct.unpack(form, line)

      Mstamp = LineCont[7]

      if Mstamp != 33686018 :
         print('input file has been created in a non-compatible computer')
         print('machine stamp is ',Mstamp,' (see imagic_format.pdf)')
         exit()

      projections.append([LineCont[3],LineCont[4],LineCont[5]])

      Nblocks = LineCont[1]
#      if Nblocks > 1 :
#         for ib in range(Nblocks-1) :
#             line = nfproj.read(nbytes)

    nfproj.close()

    FileStructN = (0,0,0)

    return projections,FileStructN

#============================
def FindKey1(nfproj,KeyWord) :

    NumTitle = 0
    while (True) :
       line = nfproj.readline()
       if not line :
          NumTitle = -1
          break

       NumTitle = NumTitle + 1
       LineCont = line.split()
       kline = len(LineCont)
       if kline > 0 and LineCont[0] == KeyWord :
          break
    
    return NumTitle
                         
#============================
def FindKey2(nfproj,KeyWord) :

    NumTitle = 0
    while (True) :
       line = nfproj.readline()
       if not line :
          NumTitle = -1
          break
       NumTitle = NumTitle + 1
       LineCont = line.split()
       kline = len(LineCont)
       if kline > 0 :
          if LineCont[0][0] == '#' or LineCont[0][0] == ';' :
             continue
          elif LineCont[0] == KeyWord :
             break
          else :
             NumTitle = -1
             break

    return NumTitle
                         
#============================
def GetLabels(nfproj) :

    NumTitle,MaxLabel              = 0, 0
    NumPosPsi,NumPosRot,NumPosTilt = 0, 0, 0

    while (True) :
      line = nfproj.readline()
      if not line :
         NumTitle = -1
         break

      LineCont = line.split()
      kline = len(LineCont)

      if kline == 0 :
         NumTitle = -1
         break

      NumTitle = NumTitle + 1
      if LineCont[0][0] == '#' or LineCont[0][0] == ';':
         continue
      elif kline <= 1 :
         NumTitle = -1
         break
      elif LineCont[0][0] != '_' :
         NumTitle = NumTitle - 1
         break
      else :
         mline = len(LineCont[0])
         if mline < 5 or LineCont[0][1] != 'r' or LineCont[0][2] != 'l' or LineCont[0][3] != 'n':
            NumTitle = -1
            break
         NumPos = int(LineCont[1].strip('#'))
         if NumPos > MaxLabel :
            MaxLabel = NumPos
         if   LineCont[0] == '_rlnAnglePsi'  :
              NumPosPsi = NumPos - 1
         elif LineCont[0] == '_rlnAngleRot'  :
              NumPosRot  = NumPos - 1
         elif LineCont[0] == '_rlnAngleTilt' :
              NumPosTilt = NumPos - 1
         else :
              continue

    NumPosAngles = (NumPosPsi,NumPosRot,NumPosTilt)

    return NumTitle,MaxLabel,NumPosAngles,line
                       
#============================
def GetDataProj(nfproj,MaxLabel,NumPosAngles,line) :

    projections = []

    NumPosPsi,NumPosRot,NumPosTilt = NumPosAngles
    NumLines = 0

#   first projections already read and read next projections

    while True :

       LineCont = line.split()
       kline = len(LineCont)

       if kline == 0 :
          break
       elif LineCont[0][0] == '#' or LineCont[0][0] == ';' :
          NumLines = NumLines + 1
#
#      line may end up by some extra symbols like '/', '\' etc making kline > MaxLabel
#
       elif kline >= MaxLabel :
          projections.append([LineCont[NumPosPsi],LineCont[NumPosRot],LineCont[NumPosTilt]])
          NumLines = NumLines + 1
       else :
          NumLines = -1
          return projections,NumLines,line

#      next projection

       line = nfproj.readline()
       if not line :
          line = 'EOF'
          break

    return projections,NumLines,line
                         
#============================
def ProjectionsViewsRelion(projections,ShowAxis,nfmes): 

    Nviews = len(projections)

    views = [ [0, 0, 0] for i in range(Nviews)]

    degrad = math.pi/180.

#   additional rotation to put the view axis normal to the page

    viewtheta, viewphi = ShowAxis
    if viewtheta == 0.0 and viewphi == 0.0 :
       FlagRot = 0
    else :
       FlagRot = 1
       viewtheta = viewtheta * degrad
       viewphi   = viewphi   * degrad
       RotView = RotTotal(RotZ,viewphi,RotY,-viewtheta,RotZ,-viewphi)

    for iv in range(Nviews):
       psi  = float(projections[iv][0]) * degrad
       rot  = float(projections[iv][1]) * degrad
       tilt = float(projections[iv][2]) * degrad

#      matrix of the rotation operator of the object is RotZ(-psi)RotY(-tilt)RotZ(-phi)
#      matrix of the inverse operator to align a projection with the object is
#      RotZ(phi)RotY(tilt)RotZ(psi) with the last column as the "views"

       RotMatrix = RotTotal(RotZ,rot,RotY,tilt,RotZ,psi)

       xv,yv,zv = (RotMatrix[2],RotMatrix[5],RotMatrix[8])

       if FlagRot > 0 :
          RotVect  = MatVect(RotView,xv,yv,zv)
          xv,yv,zv = RotVect

       if zv >= 0.0:
          views[iv] = [xv,yv,zv]
       else:
          views[iv] = [-xv,-yv,-zv]

    return views         
                         
#============================
def ProjectionsViewsImagic(projections,ShowAxis,nfmes): 

    Nviews = len(projections)

    views = [ [0, 0, 0] for i in range(Nviews)]

    degrad = math.pi/180.

#   additional rotation to put the view axis normal to the page

    viewtheta, viewphi = ShowAxis
    if viewtheta == 0.0 and viewphi == 0.0 :
       FlagRot = 0
    else :
       FlagRot = 1
       viewtheta = viewtheta * degrad
       viewphi   = viewphi   * degrad
       RotView = RotTotal(RotZ,viewphi,RotY,-viewtheta,RotZ,-viewphi)

    for iv in range(Nviews):
       alpha = float(projections[iv][0]) * degrad
       beta  = float(projections[iv][1]) * degrad
       gamma = float(projections[iv][2]) * degrad

#      matrix of the rotation operator of the object is RotZ(-alpha)RotX(-beta)RotZ(-gamma)
#      matrix of the inverse operator to align a projection with the object is
#      RotZ(gamma)RotX(beta)RotZ(alpha) with the last column as the "views"

       RotMatrix = RotTotal(RotZ,gamma,RotX,beta,RotZ,alpha)

       xv,yv,zv = (RotMatrix[2],RotMatrix[5],RotMatrix[8])

       if FlagRot > 0 :
          RotVect  = MatVect(RotView,xv,yv,zv)
          xv,yv,zv = RotVect

       if zv >= 0.0:
          views[iv] = [xv,yv,zv]
       else:
          views[iv] = [-xv,-yv,-zv]

    return views     
                             
#============================
def RotTotal(Rot1,kappa1,Rot2,kappa2,Rot3,kappa3): 

    rotmat1, rotmat2, rotmat3 = Rot1(kappa1), Rot2(kappa2), Rot3(kappa3)

    rotmat23  = MultMat(rotmat2,rotmat3)
    RotMatrix = MultMat(rotmat1,rotmat23)

    return RotMatrix     
                             
#============================
def RotX(kappa): 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] = 1.0 , 0.0 , 0.0
    rotation[1] , rotation[4] , rotation[7] = 0.0 ,  ck , -sk
    rotation[2] , rotation[5] , rotation[8] = 0.0 ,  sk ,  ck 

    return rotation     
                             
#============================
def RotY(kappa): 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] =  ck , 0.0 ,  sk
    rotation[1] , rotation[4] , rotation[7] = 0.0 , 1.0 , 0.0
    rotation[2] , rotation[5] , rotation[8] = -sk , 0.0 ,  ck 

    return rotation     
                             
#============================
def RotZ(kappa): 

    rotation = [0 for j in range(9)] 

    ck, sk = math.cos(kappa), math.sin(kappa)

    rotation[0] , rotation[3] , rotation[6] =  ck , -sk , 0.0
    rotation[1] , rotation[4] , rotation[7] =  sk ,  ck , 0.0
    rotation[2] , rotation[5] , rotation[8] = 0.0 , 0.0 , 1.0 

    return rotation     
                              
#============================
def MatVect(rot,x,y,z): 

    RotVect = [0 for i in range(3)]

    RotVect[0] = rot[0]*x + rot[3]*y + rot[6]*z
    RotVect[1] = rot[1]*x + rot[4]*y + rot[7]*z
    RotVect[2] = rot[2]*x + rot[5]*y + rot[8]*z

    return RotVect
                                                                                  
#============================
def MultMat(rot1,rot2): 

    rot3 = [0 for i in range(9)]

    rot3[0] = rot1[0]*rot2[0] + rot1[3]*rot2[1] + rot1[6]*rot2[2]
    rot3[1] = rot1[1]*rot2[0] + rot1[4]*rot2[1] + rot1[7]*rot2[2]
    rot3[2] = rot1[2]*rot2[0] + rot1[5]*rot2[1] + rot1[8]*rot2[2]
    rot3[3] = rot1[0]*rot2[3] + rot1[3]*rot2[4] + rot1[6]*rot2[5]
    rot3[4] = rot1[1]*rot2[3] + rot1[4]*rot2[4] + rot1[7]*rot2[5]
    rot3[5] = rot1[2]*rot2[3] + rot1[5]*rot2[4] + rot1[8]*rot2[5]
    rot3[6] = rot1[0]*rot2[6] + rot1[3]*rot2[7] + rot1[6]*rot2[8]
    rot3[7] = rot1[1]*rot2[6] + rot1[4]*rot2[7] + rot1[7]*rot2[8]
    rot3[8] = rot1[2]*rot2[6] + rot1[5]*rot2[7] + rot1[8]*rot2[8]

    return rot3
                                                         
#============================
def GridCenters (AngGrid,nfmes): 

    AngGrid1  = AngGrid + 1

    CellsXYZ = [ [ [0, 0, 0] for j in range(AngGrid*8+1)] for i in range(AngGrid1)]

    Uxyz1 = [0 for i in range(AngGrid1)]
    Vxyz1 = [0 for i in range(AngGrid1)]
    Uxyz2 = [0 for i in range(AngGrid1)]
    Vxyz2 = [0 for i in range(AngGrid1)] 

    for mg in range(1,AngGrid1):

        m2,  m4,  m6  = mg * 2, mg * 4, mg * 6
        m21, m41, m61 = m2 + 1, m4 + 1, m6 + 1       
        m81           = mg * 8 +1

        Rm = math.sqrt(2.) * float(mg) / float(AngGrid)
        for kg in range(mg+1):
            phi       = kg * math.pi / (4. * float(mg))
            Uxyz2[kg], Vxyz2[kg] = Rm * math.cos(phi), Rm * math.sin(phi)

        for k1 in range(mg):
            k2 = k1 + 1
            Xmk, Ymk  = (Uxyz1[k1] + Uxyz2[k2])/2. , (Vxyz1[k1] + Vxyz2[k2])/2.
            Rmk2 = (Xmk*Xmk + Ymk*Ymk) / 2.
            scalexy  = math.sqrt(1. - Rmk2/2.)
            x_sphere , y_sphere  = Xmk * scalexy, Ymk * scalexy
            z_sphere = 1.0 - Rmk2

            CellsXYZ[mg][ k2]     = [ x_sphere,  y_sphere, z_sphere]
            CellsXYZ[mg][ k2+m2]  = [-y_sphere,  x_sphere, z_sphere]
            CellsXYZ[mg][ k2+m4]  = [-x_sphere, -y_sphere, z_sphere]
            CellsXYZ[mg][ k2+m6]  = [ y_sphere, -x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m21] = [ y_sphere,  x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m41] = [-x_sphere,  y_sphere, z_sphere]
            CellsXYZ[mg][-k2+m61] = [-y_sphere, -x_sphere, z_sphere]
            CellsXYZ[mg][-k2+m81] = [ x_sphere, -y_sphere, z_sphere]

        for kg in range(mg+1):
            Uxyz1[kg], Vxyz1[kg] = Uxyz2[kg], Vxyz2[kg]

    return CellsXYZ
                                                         
#============================
def ViewsParameters (views,AngGrid,nfmes): 

    Nviews = len(views)

    viewsR = [0 for i in range(Nviews)]
    viewsA = [0 for i in range(Nviews)]
    viewsM = [0 for i in range(Nviews)]
    viewsK = [0 for i in range(Nviews)]

    pi2 = math.pi * 2.0

    for iv in range(Nviews):

        xv,yv,zv = views[iv]
       
        radv = math.sqrt(2.*(1.-zv))
        if radv > 0 :
           angv = np.arctan2(yv,xv)
           if angv < 0.0 :
              angv = angv + pi2
        else:
           angv = 0.0 
  
        mv = int(radv * float(AngGrid) / math.sqrt(2.0)) + 1
        kv = int(4.0 * angv * mv / math.pi) + 1

        viewsM [iv], viewsK [iv] = mv  , kv
        viewsR [iv], viewsA [iv] = radv, angv

    return viewsR,viewsA,viewsM,viewsK                                              

#============================
def CheckViews (views,viewsA,viewsM,AngMax,AngSigma,CellsXYZ,AngGrid,nfmes):

    Nviews = len(views)

    AngGrid1  = AngGrid     + 1
    AngGrid21 = AngGrid * 2 + 1
    AngGrid81 = AngGrid * 8 + 1

    dotmin = math.cos(AngMax)

    mdel    = int(float(AngGrid) * AngMax * 2.0 * math.sqrt(2.0) / math.pi)
    kdel    = int(float(AngGrid) * AngMax * 4.0  / math.pi)
    Mpole   = mdel * 2
    Mcircle = AngGrid - mdel

    frequencies = [ [0 for j in range(AngGrid81)] for i in range(AngGrid1)]

    for iv in range(Nviews):

        mv, av   = viewsM[iv], viewsA[iv]
        xv,yv,zv = views [iv]

        convkg   = 4.0 * av / math.pi
        convkgi  = convkg + 4

#       view close to the pole ; check the whole "cup"

        if mv <= mdel:

           for mg in range(1,Mpole+1):
               for kg in range(1, mg*8 + 1):
                   frequencies[mg][kg] += Contribution(xv,yv,zv,CellsXYZ[mg][kg],dotmin,AngSigma)

#       regular view ; check a sperical "rectangle"

        elif mv <= Mcircle:
           m1, m2 = mv - mdel, mv + mdel
           for mg in range(m1,m2):
               mg8 =  mg * 8
               kmg = int(convkg * mg) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               if k2-k1 > mg8:
                  k2 = k1 + mg8
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + mg8
                   elif kg > mg8:
                      kgt = kg - mg8
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)  

#       near-equatorial view ; two opposite views may contribute 

        else:
           m1 = mv - mdel
           for mg in range(m1,AngGrid+1):
               mg8 =  mg * 8
               kmg = int(convkg * mg) + 1 
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + mg8
                   elif kg > mg8:
                      kgt = kg - mg8
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xv,yv,zv,CellsXYZ[mg][kgt],dotmin,AngSigma)
           xvi, yvi, zvi = -xv, -yv, -zv 
           m2  = -mv + AngGrid21 - mdel
           for mg in range(m2,AngGrid+1):
               mg8 =  mg * 8
               kmg = int(convkgi * mg) + 1
               k1, k2 = kmg - kdel, kmg + kdel
               for kg in range(k1,k2+1):
                   if kg <= 0:
                      kgt = kg + mg8
                   elif kg > mg8:
                      kgt = kg - mg8
                   else:
                      kgt = kg
                   frequencies[mg][kgt] += Contribution(xvi,yvi,zvi,CellsXYZ[mg][kgt],dotmin,AngSigma)   

    factnorm  = 1. / (2. * math.pi * AngSigma)
    unitsurf  = math.pi / ( 2. * float(AngGrid*AngGrid))
    scalenorm = factnorm * unitsurf / float(Nviews)

    for mg in range(1,AngGrid1):
        for kg in range(1,mg*8+1):
            frequencies[mg][kg] = frequencies[mg][kg] * scalenorm

    return frequencies

#============================
def StatFrequencies (frequencies,AngGrid,ParamStat,nfmes):

#   bound values and validation

    ControlSum, GridSum, SumNul, FreqMax, FreqMin = 0.0, 0.0, 0.0, 0.0, 1.0

    for mg in range(1,AngGrid+1):
        mg3, mg5, mg7 = mg * 3, mg * 5, mg * 7
        for kg in range(1,mg*8+1):
            if (kg != mg and kg != mg3 and kg != mg5 and kg != mg7):
               GridSum    = GridSum   + 1
               freq       = frequencies[mg][kg]
               ControlSum = ControlSum + freq
               if freq > FreqMax:
                  FreqMax = freq
               if freq <= 0.0:
                  SumNul = SumNul +1
               elif freq < FreqMin:
                  FreqMin = freq

    if SumNul == GridSum:
       print('*** NO CONTRIBUTION FOUND ***')
       print('*** Try a finer grid      ***')
       exit()

    FreqMean                 = ControlSum / (4.0 * float(AngGrid)**2 )
    FreqMinAver, FreqMaxAver = FreqMin / FreqMean ,  FreqMax / FreqMean
    FreqMinLog,  FreqMaxLog  = math.log2(FreqMinAver) , math.log2(FreqMaxAver) 

    print('',file=nfmes)
    print('============== data statistics ==============',file=nfmes)
    print('',file=nfmes)
    print(f'Mean frequency value',32*' ',f'{FreqMean:.5E}',file=nfmes)
    print(f'Minimal positive and Maximal normalized frequencies  ',
          f'{FreqMin:.5E}  {FreqMax:.5E}',file=nfmes)
    print(f'Min and Max frequencies w/respect to the mean value  ',
          f'{FreqMinAver:.5E}  {FreqMaxAver:.5E}',file=nfmes)
    print(f'Log2 Min and Max frequencies w/respect to the mean ',
          f'{FreqMinLog:8.5f}     {FreqMaxLog:9.6f}',file=nfmes)

#   histogram parameters

    HistNum,HistBottom,HistTop = ParamStat

    if HistBottom == 'unkn':
       HistBottom =  FreqMinAver
    HistBottomInt =  int(math.log2(HistBottom)) - 1
    HistBottom    =  float(HistBottomInt)

    if HistTop == 'unkn':
       HistTop =  FreqMaxAver
    HistTopInt =  int(math.log2(HistTop))
    HistTop    =  float(HistTopInt)

    if HistNum == 'unkn':
       HistNum =  HistNumC = int(HistTop-HistBottom)

    print(f'Applied number of bins for the Log2 Freq-histogram {HistNum:4}',file=nfmes)
    print(f'Applied Log2 Min/mean and Max/mean frequencies  ',
          f'{HistBottom:8.1f}     {HistTop:8.1f}',file=nfmes)
    
    Hist1    = [0 for i in range(HistNum+1)]
    Hist2    = [0 for i in range(HistNum+1)]
    CumHist1 = [0 for i in range(HistNum+1)]
    CumHist2 = [0 for i in range(HistNum+1)]
#
#   histograms calculation
#
    StepHist   = float(HistTop - HistBottom) / float(HistNum)
    SumTotal   = 4 * AngGrid**2
    bin1, bin2 = 1.0 / float(GridSum) , 1.0 / float(GridSum-SumNul)

    for mg in range(1,AngGrid+1):
        mg3, mg5, mg7 = mg * 3, mg * 5, mg * 7
        for kg in range(1,mg*8+1):
            if (kg != mg and kg != mg3 and kg != mg5 and kg != mg7):
               freq = frequencies[mg][kg]

               if freq <= 0.0:
                  ih = 0
               else :
                  freqlog = math.log2(freq/FreqMean)
                  if freqlog <= HistBottom :
                     ih = 1
                  elif freqlog >= HistTop :
                     ih = HistNum
                  else:
                     ih = int((freqlog - HistBottom) / StepHist) + 1

               Hist1[ih] = Hist1[ih] + 1

    for ih in range(HistNum+1):
        hist                 = Hist1[ih]
        Hist1[ih], Hist2[ih] = hist * bin1, hist * bin2
    Hist2[0] = 0

#   cumulative distributions

    CumHist1[0], CumHist2[0] = Hist1[0], Hist2[0]

    for ih in range(HistNum):
        CumHist1[ih+1] = CumHist1[ih] + Hist1[ih+1]
        CumHist2[ih+1] = CumHist2[ih] + Hist2[ih+1]

    ParamStatN = (HistNum,HistBottom,HistTop)

    return frequencies,ControlSum,GridSum,Hist1,Hist2,CumHist1,CumHist2,ParamStatN

#============================
def Contribution(xv,yv,zv,grid,dotmin,AngSigma):

    dotprod = grid[0]*xv + grid[1]*yv + grid[2]*zv

    freq = math.exp((dotprod-1.) / AngSigma)

#    if dotprod >= dotmin:
#       freq = math.exp((dotprod-1.) / AngSigma)
#    else:
#       freq = 0.0

    return freq

#============================
def CellsToViews (views,viewsM,viewsK,viewsR,viewsA,GridFreq,
                                            AngGrid,AngFreq,nfmes):

    Nviews = len(views)

    ViewsFreq = [0 for i in range(Nviews)]

#   view frequency = frequency of the cell

    if AngFreq == 1:
       ViewsFreq = ViewsDirect(viewsM,viewsK,GridFreq)

    elif AngFreq == 2:

#   first get interpolated frequency for the grid nodes
#   then get linear interpolations for the views frequencies

       NodesFreq = CellsToGrid(GridFreq,AngGrid,nfmes)

       ViewsFreq = GridToViews(views,viewsM,viewsK,viewsR,viewsA,
                                            NodesFreq,AngGrid,nfmes)
    else:
       print('*** No method to calculate individual views frequencies is chosen',
              file=nfmes)
       exit ()

#   normalize views frequencies

    SumFreqViews = 0.0
    for iv in range(Nviews):
        SumFreqViews = SumFreqViews + ViewsFreq[iv]

    scale = float(Nviews) / SumFreqViews
    SumFreqViews = 0.0
    for iv in range(Nviews):
        ViewsFreq[iv] = ViewsFreq[iv] * scale
        SumFreqViews  = SumFreqViews  + ViewsFreq[iv]

    return ViewsFreq,SumFreqViews
  
#============================
def ViewsDirect(viewsM,viewsK,GridFreq):

    Nviews = len(viewsM)

    ViewsFreq = [0 for i in range(Nviews)]

    for iv in range (Nviews):
        mv, kv        = viewsM[iv], viewsK[iv]
        ViewsFreq[iv] = GridFreq[mv][kv]

    return ViewsFreq

#============================
def CellsToGrid (GridFreq,AngGrid,nfmes):

    NodesFreq = [ [0 for j in range(AngGrid*8+1)] for i in range(AngGrid+1)]

#   pole and axis k = 0 ; no equator

    NodesFreq [0][0] = (GridFreq[1][1] + GridFreq[1][2] +
                        GridFreq[1][3] + GridFreq[1][4]) /4.

    for m in range(1,AngGrid):
        m1, m8 = m+1 , m*8
        NodesFreq [m][0 ] = (GridFreq[m][1]  + GridFreq[m][m8] + 
                             GridFreq[m1][1] + GridFreq[m1][m8+8]) /4.
        NodesFreq [m][m8] = NodesFreq [m][0 ]

#   diagonals with triangular cells k = m , k = 3m , k = 5m , k = 7m ; no equator

    for m in range(1,AngGrid):
        m1, m3, m5, m7    = m + 1, m * 3, m * 5, m * 7
        NodesFreq [m][m]  = (GridFreq[m][m]  + GridFreq[m1][m1]  ) /2.
        NodesFreq [m][m3] = (GridFreq[m][m3] + GridFreq[m1][m3+3]) /2.
        NodesFreq [m][m5] = (GridFreq[m][m5] + GridFreq[m1][m5+5]) /2.
        NodesFreq [m][m7] = (GridFreq[m][m7] + GridFreq[m1][m7+7]) /2.

#   internal sectors

    for m in range(1,AngGrid):
        m1 = m+1
        for k in range(1,m):
            NodesFreq [m][k]  = (GridFreq[m][k]    + GridFreq[m][k+1]+
                                 GridFreq[m1][k]   + GridFreq[m1][k+1]) /4.
        for k in range(m1,m3):
            NodesFreq [m][k]  = (GridFreq[m][k]    + GridFreq[m][k+1]+
                                 GridFreq[m1][k+2] + GridFreq[m1][k+3]) /4.
        for k in range(m3+1,m5):
            NodesFreq [m][k]  = (GridFreq[m][k]    + GridFreq[m][k+1]+
                                 GridFreq[m1][k+4] + GridFreq[m1][k+5]) /4.
        for k in range(m5+1,m7):
            NodesFreq [m][k]  = (GridFreq[m][k]    + GridFreq[m][k+1]+
                                 GridFreq[m1][k+6] + GridFreq[m1][k+7]) /4.
        for k in range(m7+1,m*8):
            NodesFreq [m][k]  = (GridFreq[m][k]    + GridFreq[m][k+1]+
                                 GridFreq[m1][k+8] + GridFreq[m1][k+9]) /4.
  
#   equator

    M4, M8 = AngGrid * 4 , AngGrid * 8
    for k in range(1,M4):
        NodesFreq [AngGrid][k]    = (GridFreq[AngGrid][k]    + GridFreq[AngGrid][k+1]+
                                     GridFreq[AngGrid][k+M4] + GridFreq[AngGrid][k+M4+1])/4.
        NodesFreq [AngGrid][k+M4] = NodesFreq [AngGrid][k]
       
    NodesFreq [AngGrid][0]  = (GridFreq[AngGrid][M8] + GridFreq[AngGrid][1]+
                               GridFreq[AngGrid][M4] + GridFreq[AngGrid][M4+1])/2.
    NodesFreq [AngGrid][M4] = NodesFreq[AngGrid][0]
    NodesFreq [AngGrid][M8] = NodesFreq[AngGrid][0]

    return NodesFreq
  
#============================
def GridToViews(views,viewsM,viewsK,viewsR,viewsA,NodesFreq,AngGrid,nfmes):

    Nviews = len(views)

    ViewsFreq = [0 for i in range(Nviews)]

    scaler = float(AngGrid) / math.sqrt(2.0)
    scalea = 4.0 / math.pi

    for iv in range(Nviews):
        mv, kv = viewsM[iv], viewsK[iv]
        rv, av = viewsR[iv], viewsA[iv]

        rm   = rv * scaler
        aint = mv - rm

#       interpolation by radius, between the points A-C and B-D

        FAC = NodesFreq[mv-1][kv-1] * aint + NodesFreq[mv][kv-1] * (1.0-aint)
        FBD = NodesFreq[mv-1][kv  ] * aint + NodesFreq[mv][kv  ] * (1.0-aint)

#       interpolation by angle between interpolated (AC) and (BD)

        ak   = av * rm * scalea
        bint = kv - ak

        ViewsFreq[iv] = FAC * bint + FBD * (1.0-bint)

    return ViewsFreq
  
#============================
def DrawAll(CellsFreq,AngGrid,viewsR,viewsA,ControlSum,ViewsFreq,
            Hist1,Hist2,CumHist1,CumHist2,ParamStat,ParamFig,ParamPlot,ParamAux,nfmes):
    
    Nviews = len(viewsR)

    HistNum,HistBottom,HistTop                                            = ParamStat
    DrawHist,DrawMap,DrawView,SizeX,ShiftH,ShowDots                       = ParamFig
    ColorHist,ColorPlus,ColorMap,ColorViews,ColorBottom,ColorTop,ColorNum = ParamPlot
    ParamAuxBar,TitleX,ParamAuxHist,ParamAuxSize                          = ParamAux

    ColorBox1,ColorBox2,ColorBox4,NumBarTicks = ParamAuxBar
    HistBarWidth,HistBarShift                 = ParamAuxHist
    SizeLabMin, SizeLabMax, SizeLabScale      = ParamAuxSize

    FreqLog1 = [0 for i in range(HistNum+1)]
    FreqLog2 = [0 for i in range(HistNum+1)]

    if DrawHist == 'yes' :
       FlagH =  1
    else :
       FlagH =  0

    if DrawMap == 'yes' :
       FlagM =  1
    else :
       FlagM =  0

    if DrawView == 'yes' :
       FlagV =  1
    else :
       FlagV =  0 

#   color scale with respect to the mean frequency in the log2-scale 
#   map from (ColorBottom , ColorTop) to (0 , ColorNum)

    ColorTopL2    = math.log(float(ColorTop)    , 2.0)
    ColorBottomL2 = math.log(float(ColorBottom) , 2.0)

    ColorStep = float(ColorNum) / (float(ColorTopL2 - ColorBottomL2))
    FreqMean  = ControlSum / (4.0 * float(AngGrid)**2 )

#   colors and the colorbar

    colors    = cm.get_cmap(ColorMap)
    newcolors = colors(np.linspace(0, 1, ColorNum+1))

    ColorScale = mpl.colors.Normalize(vmin=ColorBottomL2, vmax=ColorTopL2)
    ColorBar   = plt.cm.ScalarMappable(cmap=colors, norm=ColorScale)
    ColorBar.set_array([])

#   define the label size

    SizeLab = SizeX * 2.54
    if SizeLab < SizeLabMin :
       SizeLab = SizeLabMin
    elif SizeLab > SizeLabMax : 
       SizeLab = SizeLabMax
    LabelSizeA, LabelSizeB = int(SizeLab),              int(SizeLab)
    LabelSizeT, LabelSizeF = int(SizeLab*SizeLabScale), int(SizeLab*SizeLabScale)

    SizeY   = SizeX
    SizeYH  = SizeY * 0.95
    SizeX2  = 2.0 * SizeX + ShiftH
    SizeY2  = 2.0 * SizeY

#   define the figure composition ; prepare the frames for images

    title_H = 'Histogram of log2(Freq/Aver)'
    title_C = 'Cumulative function of log2(F/A)'
    title_D = 'Distribution map'
    title_V = 'Views distribution'

    Nmaps   = FlagM + FlagV

    if FlagH + Nmaps <= 0 :
       return

    if FlagH > 0 : 
       if Nmaps == 0 :
          fig, (axe1,axe2) = plt.subplots(1,2,figsize = (SizeX2,SizeYH))
          axe1.set_title(title_H,fontsize=LabelSizeT)
          axe2.set_title(title_C,fontsize=LabelSizeT)

       else :
          fig, ((axe3,axe4),(axe1,axe2)) = plt.subplots(2,2,figsize = (SizeX2,SizeY2))
          axe1.set_title(title_H,fontsize=LabelSizeT)
          axe2.set_title(title_C,fontsize=LabelSizeT)
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)
          ColorBarBox = ColorBox4

    else:
       ColorBarBox = ColorBox1
       if Nmaps == 2 :
          fig, (axe3,axe4) = plt.subplots(1,2,figsize = (SizeX2,SizeY))
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)
          ColorBarBox = ColorBox2
       elif FlagM == 1 :
          fig, (axe3) = plt.subplots(1,1,figsize = (SizeX,SizeY))
          axe3.set_title(title_D,fontsize=LabelSizeT)
          axe3.get_xaxis().set_visible(False)
          axe3.get_yaxis().set_visible(False)
       else :
          fig, (axe4) = plt.subplots(1,1,figsize = (SizeX,SizeY))
          axe4.set_title(title_V,fontsize=LabelSizeT)
          axe4.get_xaxis().set_visible(False)
          axe4.get_yaxis().set_visible(False)

#   plot colorbar

    if Nmaps > 0 :
       mpl.rcParams['ytick.labelsize'] = LabelSizeB
       CBar = fig.add_axes(ColorBarBox)
       BarTicks=np.linspace(ColorBottomL2, ColorTopL2, NumBarTicks)
       plt.colorbar(ColorBar, cax=CBar, orientation='vertical',ticks=BarTicks)
       CBar.set_yticklabels(['{:4.1f}'.format(i) for i in BarTicks])

#   plot main title

    fig.suptitle('VUE',fontsize=LabelSizeF,fontweight='bold',x=TitleX)

#   plot the frequency histogram and the cumulative distribution

    dh = float(HistTop - HistBottom) / float(HistNum)

    for ih in range(HistNum+1) :
        FreqLog1[ih] = dh * ih + HistBottom
        FreqLog2[ih] = dh * ih + HistBottom + HistBarShift

    if FlagH > 0 :
       axe1.tick_params(axis='x', labelsize=LabelSizeA)
       axe1.tick_params(axis='y', labelsize=LabelSizeA)
       axe2.tick_params(axis='x', labelsize=LabelSizeA)
       axe2.tick_params(axis='y', labelsize=LabelSizeA)
       axe1.bar(FreqLog1,Hist1,   color=ColorHist,width=HistBarWidth, align='edge')
       axe1.bar(FreqLog2,Hist2,   color=ColorPlus,width=HistBarWidth, align='edge')
       axe2.bar(FreqLog1,CumHist1,color=ColorHist,width=HistBarWidth, align='edge')
       axe2.bar(FreqLog2,CumHist2,color=ColorPlus,width=HistBarWidth, align='edge')

#   plot distribution

    if(FlagM > 0):

       axe3.add_artist(
                ptc.Circle((0.5,0.5), 0.5, facecolor='white',edgecolor='black', zorder = 1))

       AngGrid8 = AngGrid * 8

       for m in range(1,AngGrid+1):
           rad2, rad1 = float(m)/float(AngGrid) , float(m-1)/float(AngGrid)
           delang2 = math.pi / float(m*4) 
           if m > 1 :
              delang1 = math.pi / (float(m-1)*4) 
           else :
              delang1 = 0.0

           for k2 in range(1,m*8+1):
               if k2 != m or k2 != m*3 or k2 != m*5 or k2 != m*7 :
                  freq      = CellsFreq[m][k2]
                  CellColor = GetColor(freq,FreqMean,ColorBottomL2,ColorTopL2,ColorStep,
                                       newcolors,ColorNum)
                  coord     = GetCell(m,k2,rad1,rad2,delang1,delang2)

                  axe3.add_artist(ptc.Polygon(coord, closed=True, color = CellColor))

#   draw grid
#                  axe3.add_artist(ptc.Polygon(coord, closed=True,
#                                 facecolor='white',edgecolor='black',linewidth=0.5))
#
#   plot individual views

    if FlagV > 0:

      axe4.add_artist(
               ptc.Circle((0.5,0.5), 0.5, facecolor='white',edgecolor='black', zorder = 1))

      ViewMean  = 1.0

      for iv in range(Nviews):
          rad = viewsR[iv] / math.sqrt(2.0)
          ang = viewsA[iv]
          xv, yv = (rad * math.cos(ang)) * 0.5 + 0.5 , (rad * math.sin(ang)) * 0.5 + 0.5

          if ColorViews > 0 :
              freq     =  ColorTop
          elif ColorViews < 0 :
              freq     =  ColorBottom
          else :
              freq     = ViewsFreq[iv]

          ViewColor = GetColor(freq,ViewMean,ColorBottomL2,ColorTopL2,ColorStep,
                               newcolors,ColorNum)

          axe4.add_artist(ptc.Circle((xv,yv), ShowDots, color = ViewColor, zorder = 2))

    plt.show()

    return 
  
#============================
def GetCell(m,k2,rad1,rad2,delang1,delang2):

    coord  = [ [0,0] for i in range(4)]

    k1 = k2
    if k2 > m :
       k1 = k1 - 2
    if k2 > m*3 :
       k1 = k1 - 2
    if k2 > m*5 :
       k1 = k1 - 2
    if k2 > m*7 :
       k1 = k1 - 2

    angle0 = delang2 * k2
    angle3 = delang2 * (k2-1)
    anglem = delang2 * (k2-2)
    angle1 = delang1 * k1
    angle2 = delang1 * (k1-1)

    coord[0] = [rad2 * math.cos(angle0) , rad2 * math.sin(angle0)]
    coord[3] = [rad2 * math.cos(angle3) , rad2 * math.sin(angle3)]
    coord[1] = [rad1 * math.cos(angle1) , rad1 * math.sin(angle1)]

    if k2 == m+1 or k2 == m*3+1 or k2 == m*5+1 or k2 == m*7+1 :
       coord[2] = [rad2 * math.cos(anglem) , rad2 * math.sin(anglem)]
    else :
       coord[2] = [rad1 * math.cos(angle2) , rad1 * math.sin(angle2)]
       
    for i in range(4):
        coord[i][0] = coord[i][0] * 0.5 + 0.5
        coord[i][1] = coord[i][1] * 0.5 + 0.5

    return coord

#============================
def GetColor(freq,FreqMean,ColorBottom,ColorTop,ColorStep,newcolors,ColorNum):
    if freq > 0.0:
       ColorValue  = math.log(freq/FreqMean , 2.0)
       if   ColorValue >= ColorTop :
            CellColor  =  newcolors[ColorNum]
       elif ColorValue <= ColorBottom :
            CellColor  =  newcolors[0]
       else:
            CellColor = newcolors[int((ColorValue - ColorBottom) * ColorStep)]
    else :
       CellColor = newcolors[0]

    return CellColor

#============================
def PrintStat(frequencies,AngGrid,ControlSum,GridSum,SumFreqViews,
                       Hist1,Hist2,CumHist1,CumHist2,ParamStat,PrintFlag,nfmes):

    print('',file=nfmes)
    print('============== procedure report ==============',file=nfmes)

#   print views statistics

    print('',file=nfmes)
    print(f'Total number of the grid cells',' '*12,f'{int(GridSum):11}'       ,file=nfmes)
    print(f'Sum of the views frequencies',' '*17,f'{SumFreqViews:15.6f}'      ,file=nfmes)
    print(f'Sum of the normalized cells frequencies        {ControlSum:15.6f}',file=nfmes)

    HistNum,HistBottom,HistTop = ParamStat

#   print histograms of the grid values 

    print('',file=nfmes)
    print('   n  Log2(F)     Frequency   Hist_All  Hist_Fill  Cumul_All  Cumul_Fill',
                                                                               file=nfmes)

    StepHist   = float(HistTop - HistBottom) / float(HistNum)
 
    if PrintFlag > 0:
       for ih in range(HistNum+1) :
           Log2Freq = HistBottom+ih*StepHist
           freq     = 2.0**Log2Freq
           print(f'{ih:4} {Log2Freq:7.2f}   {freq:12.6f} {Hist1[ih]:10.6f} {Hist2[ih]:10.6f}',
                 f'{CumHist1[ih]:10.6f} {CumHist2[ih]:10.6f}',
                 file=nfmes)

    print('',file=nfmes)

#   print grid values
 
    if PrintFlag > 1:
       print('',file=nfmes)
       for mg in range(AngGrid+1):
           for kg in range(mg*8+1):
               if frequencies[mg][kg] > 0:
                  print(f'grid cell (AngGrid,kgrid)= ({mg:3},{kg:3})',  
                        f'    frequency={frequencies[mg][kg]:8.5f}',file=nfmes)

    return

#============================
def OutViews(views,ParamFileN,AngWeight): 

    ParamFile,FileStructN                       = ParamFileN
    nfmes,PrintFlag,FileProj,FileProjW,FileType = ParamFile

    if FileProjW == 'unkn' :
       return

#   calculate weights

    if AngWeight == 1 :
       weights = ViewsToWeights1(views)
    elif AngWeight == 2 :
       weights = ViewsToWeights2(views)
    else :
       weights = ViewsToWeights0(views)

#   output calculated frequencies and weights
#   check, and correct if necessary, the file-name extension

    lnfile = len(FileProjW)

    if FileType == 'test' :
       WriteTest(views,weights,FileProj,FileProjW,FileStructN)

    elif FileType == 'star' :
       if lnfile < 6 or FileProjW[lnfile-5:lnfile] != '.star' :
          FileProjW = FileProjW + '.star'
       WriteStar(views,weights,FileProj,FileProjW,FileStructN)

    else :
       if lnfile < 5 or FileProjW[lnfile-4:lnfile] != '.hed' :
          FileProjW = FileProjW + '.hed'
       WriteHed(views,weights,FileProj,FileProjW,FileStructN)

    return

#============================
def ViewsToWeights0(views): 

    Nviews = len(views)

    weights = [0.0 for i in range(Nviews)]

    return weights


#============================
def ViewsToWeights1(views): 

    Nviews = len(views)

    weights = [1./views[i] for i in range(Nviews)]

    return weights


#============================
def ViewsToWeights2(views): 

    Nviews = len(views)

    alog2 = math.log(2.0)

    weights = [(math.log(views[i]+1.0)/alog2) / views[i] for i in range(Nviews)]

    return weights

#============================
def WriteTest(views,weights,FileProj,FileProjW,FileStructN): 

    nfread  = open(FileProj , 'r')
    nfwrite = open(FileProjW, 'w')

    Nviews  = len(views)

    for iv in range(Nviews) :
        line    = nfread.readline()
        LineCut = line.rstrip('\n')
        print(f'{LineCut} {views[iv]:12.7f} {weights[iv]:12.7f}',file=nfwrite)

    return

#============================
def WriteStar(views,weights,FileProj,FileProjW,FileStructN): 

    NumTitles,MaxLabel,NumPosAngles             = FileStructN 

    nfread  = open(FileProj , 'r')
    nfwrite = open(FileProjW, 'w')

    Nviews  = len(views)

#   copy headers including old label descriptions

    for i in range (NumTitles) :
       line    = nfread.readline()
       LineCut = line.rstrip('\n')
       print(LineCut,file=nfwrite)
          
#   add two new label descriptions

    if MaxLabel < 10 :
       print(f'_rlnViewsFrequency #{MaxLabel+1:1}',file=nfwrite)
    elif MaxLabel < 100 :
       print(f'_rlnViewsFrequency #{MaxLabel+1:2}',file=nfwrite)
    else :
       print(f'_rlnViewsFrequency #{MaxLabel+1:3}',file=nfwrite)

    if MaxLabel < 9 :
       print(f'_rlnViewsWeights #{MaxLabel+2:1}',file=nfwrite)
    elif MaxLabel < 99 :
       print(f'_rlnViewsWeights #{MaxLabel+2:2}',file=nfwrite)
    else :
       print(f'_rlnViewsWeights #{MaxLabel+2:3}',file=nfwrite)

#   write projections

    iv = 0 
    
    while True :
       line  = nfread.readline()
       kline = len(line)
       if kline == 0 or line[0] == '#' or line[0] == ';' :
          print(line,file=nfwrite)
       else :
          LineCut = line.rstrip('\n')
          print(f'{LineCut} {views[iv]:12.7f} {weights[iv]:12.7f}',file=nfwrite)
          iv = iv + 1
          if iv == Nviews :
             break

#   copy eventual last records

    while True :
       line = nfread.readline()
       if not line :
          break
       else :
          LineCut = line.rstrip('\n')
          print(LineCut,file=nfwrite)

    nfread.close()

    return

#============================
def WriteHed(views,weights,FileProj,FileProjW,FileStructN): 

    nfproj  = open(FileProj , 'rb')
    nfwrite = open(FileProjW, 'wb')

#   detailed Imagic format (not required here)
#    form = '@14i 4s 2i 12f 80s f 14i 3f 31i 3f 2i 3f i 13f 2i 77f 228s'

#   format to add multiplicity and weight
    form_in  = '@12s i 232s 4s 4s 768s'
    form_out = '@12s i 232s  f  f 768s'

    nbytes = 1024

    Nviews  = len(views)

    for iv in range(Nviews) :
        line_in = nfproj.read(nbytes)
        LineCont = struct.unpack(form_in, line_in)

        line_out = struct.pack(form_out,LineCont[0],LineCont[1],LineCont[2],
                                        views[iv],weights[iv],LineCont[5])
        nfwrite.write(line_out)

        Nblocks = LineCont[1]
        if Nblocks > 1 :
           for ib in range(Nblocks-1) :
               line = nfproj.read(nbytes)
               nfwrite.write(line)

    nfproj.close()

    return

####################################################
#
#  MAIN PROGRAM

# Import math Library

#import math 

#
#  fixed parameters
#

filemes  = "VUE.log"
filedata = "VUE.dat"

#nfmes = open(filemes, 'w')

#==================== input information =====================

#     input parameters

ParamCalc,ParamStat,ParamFig,ParamPlot,ParamFile,ParamAux = InputData(filedata,filemes)

AngGrid,AngEps,AngMax,AngSigma,AngFreq,AngWeight = ParamCalc
nfmes, PrintFlag                                 = ParamFile[0][0], ParamFile[0][1]

#     input projection descriptions and conversion into views

views,ParamFileN = GetViews(ParamFile)

Nviews = len(views)

print(Nviews,' input views')
print('Preparating spherical grid ...')

#=============== data processing and preparation =====================

#     prepare the grids

CellsXYZ = GridCenters (AngGrid,nfmes)

#     get views parameters

viewsR,viewsA,viewsM,viewsK = ViewsParameters(views,AngGrid,nfmes)

print('       Sperical grid and views coordinates prepared')
print('Calculating frequencies in the grid ...')

#=============== cells frequencies calculation =====================

CellsFreq = CheckViews(views,viewsA,viewsM,AngMax,AngSigma,CellsXYZ,AngGrid,nfmes)

print('       Frequencies in the grid calculated')
print('Calculating statistics ...')

#====

CellsFreq,ControlSum,GridSum,Hist1,Hist2,CumHist1,CumHist2,ParamStatN           \
                             = StatFrequencies(CellsFreq,AngGrid,ParamStat,nfmes)

print('       Frequencies statistics calculated')
print('Calculating frequencies for individual views...')

#=============== views frequencies calculation =====================

ViewsFreq,SumFreqViews = CellsToViews(views,viewsM,viewsK,viewsR,viewsA,
                                                CellsFreq,AngGrid,AngFreq,nfmes)

print('       Frequencies for individual views calculated')
print('Printing statistics ...')

#=============== print statistics =====================

PrintStat(CellsFreq,AngGrid,ControlSum,GridSum,SumFreqViews,
                        Hist1,Hist2,CumHist1,CumHist2,ParamStatN,PrintFlag,nfmes)

print('       Statistics printed')
print('Output file with views weigths if required...')

#=============== output file with views' weights =====================

OutViews(ViewsFreq,ParamFileN,AngWeight)

print('       Output file with views weigths created')
print('Preparing plots if required...')

#=============== plot calculated frequencies =====================

DrawAll(CellsFreq,AngGrid,viewsR,viewsA,ControlSum,ViewsFreq,
        Hist1,Hist2,CumHist1,CumHist2,ParamStatN,ParamFig,ParamPlot,ParamAux,nfmes)

print('       Plots shown')
